'''Testing suite to verify flash-amr-tools'''
import flash_amr_tools
import numpy as np
import h5py

import os 
this_dir = os.path.dirname(os.path.realpath(__file__))

import pytest
import urllib.request

@pytest.fixture
def sedov_cube_data():
    '''Return the cubed Sedov data'''
    f = h5py.File(os.path.join(this_dir, "sedov_cube_data.h5"), "r")
    yield f
    f.close()

def get_sedov_datafile():
    '''Get sedov data from yt hub'''
    fname, _ = urllib.request.urlretrieve("https://hub.yt/api/v1/item/577c13b50d7c6b0001ad63f1/download", filename=os.path.join(this_dir,"sedov_hdf5_chk_0003"))
    return fname


def test_dims(sedov_cube_data):
    '''Check dimensionality of the cube'''

    # get data first
    filename = get_sedov_datafile()

    # get the block lists
    blist, brefs, bns = flash_amr_tools.get_true_blocks(filename)

    # read in the data
    pf = h5py.File(filename)
    dens = pf["dens"][()][blist]
    ref_lvl = pf["refine level"][()][blist]
    bbox = pf["bounding box"][()][blist]
    bsize = pf["block size"][()][blist]
    pf.close()

    # compute cube
    dens_cube = flash_amr_tools.get_cube(dens, ref_lvl=ref_lvl, bbox=bbox, bsize=bsize, brefs=brefs, bns=bns)
    
    # check the dimensionality
    assert np.all(dens_cube.shape == sedov_cube_data["dims"][()]), "Dimensions do not align!"

def test_cube(sedov_cube_data):
    '''Check cube data'''

    # get data first
    filename = get_sedov_datafile()

    # get the block lists
    blist, brefs, bns = flash_amr_tools.get_true_blocks(filename)

    # read in the data
    pf = h5py.File(filename)
    dens = pf["dens"][()][blist]
    ref_lvl = pf["refine level"][()][blist]
    bbox = pf["bounding box"][()][blist]
    bsize = pf["block size"][()][blist]
    pf.close()

    # compute cube
    dens_cube = flash_amr_tools.get_cube(dens, ref_lvl=ref_lvl, bbox=bbox, bsize=bsize, brefs=brefs, bns=bns)

    # check if data is close enough to true data
    assert np.all(np.isclose(dens_cube, sedov_cube_data["cube"][()])), "Cube data does not align!"

def test_slice(sedov_cube_data):
    '''Check slice data'''

    # get data first
    filename = get_sedov_datafile()

    # get the block lists
    blist, brefs, bns = flash_amr_tools.get_true_blocks(filename)

    # read in the data
    pf = h5py.File(filename)
    dens = pf["dens"][()][blist]
    ref_lvl = pf["refine level"][()][blist]
    bbox = pf["bounding box"][()][blist]
    bsize = pf["block size"][()][blist]
    pf.close()

    # compute slice along z=0
    dens_sl = flash_amr_tools.get_slice(dens, pos=0.5, axis=2, ref_lvl=ref_lvl, bbox=bbox, bsize=bsize, brefs=brefs, bns=bns)

    # check if data is close enough to true data
    assert np.all(np.isclose(dens_sl, sedov_cube_data["dens_sl"][()])), "Slice data does not align!"


def test_cdens(sedov_cube_data):
    '''Check column density data'''

    # get data first
    filename = get_sedov_datafile()

    # get the block lists
    blist, brefs, bns = flash_amr_tools.get_true_blocks(filename)

    # read in the data
    pf = h5py.File(filename)
    dens = pf["dens"][()][blist]
    ref_lvl = pf["refine level"][()][blist]
    bbox = pf["bounding box"][()][blist]
    bsize = pf["block size"][()][blist]
    pf.close()

    # compute column density
    dens_cdens = flash_amr_tools.get_cdens(dens, axis=2, ref_lvl=ref_lvl, bbox=bbox, bsize=bsize, brefs=brefs, bns=bns)

    # check if data is close enough to true data
    assert np.all(np.isclose(dens_cdens, sedov_cube_data["dens_cdens"][()])), "Column density data does not align!"


'''
Toolkit to convert from AMR data in FLASH -> Uniform data
'''

import numpy as np
import os, sys
import h5py

# relative imports for finding blocks & creating block lists
from .block_tools import find_blocks, create_blists


def get_true_blocks(
        fname: str,
        xmin=np.array([], dtype=np.float32),
        xmax=np.array([], dtype=np.float32),
        is_cuboid=True,
        max_ref_given=None,
        min_ref_given=None,
        verbose=True,
    ):
    '''
    Extract the complete list of blocks which are in the box chosen within xmin and xmax. In particular, it sets the minimum & maximum refinement within the block list, and the number of blocks in x, y, z on the lowest refinement level.

    - fname (str): the filename of the FLASH output file. needs to be a .h5 file
    - xmin (3-D array): the lower corner of the region of interest. Defaults to empty array, which uses the whole domain
    - xmax (3-D array): the upper corner of the region of interest. Defaults to empty array, which uses the whole domain
    - is_cuboid (bool): set to True if the region is not a cube. Defaults to True.
    - max_ref_given (int): force the maximum refinement level of the blocks. Defaults to None, which uses the maximum refinement level in the simulation.
    - min_ref_given (int): force the minimum refinement level of the blocks. Defaults to None, which uses the minimum refinement level in the simulation.
    - verbose (bool) : show print statements or not.

    Returns:
    - blist (list) : the list of blocks that cover the region of interest 
    - brefs (list) : the maximum & minimum refinement level within the region of interest
    - bns (list) : returns the number of blocks in (x, y, z) direction at the lowest refinement level
    '''

    # makes sure that the filename exists
    if not os.path.exists(fname):
        raise FileNotFoundError("File %s is not found!" % fname)

    # Read in the FLASH file and create the 3d grid of the blocks.
    pf = h5py.File(name=fname, mode='r')

    # Define some pointers for regular used data tables.
    coords = pf['coordinates'][()]
    gid = pf['gid'][()]
    refine_level = pf['refine level'][()]
    block_size = pf['block size'][()]

    # Reading the border of the simulation box.
    sim_dict = dict(pf['real runtime parameters'][()])

    simxl = sim_dict[('%-80s' % 'xmin').encode('UTF-8')]
    simxr = sim_dict[('%-80s' % 'xmax').encode('UTF-8')]
    simyl = sim_dict[('%-80s' % 'ymin').encode('UTF-8')]
    simyr = sim_dict[('%-80s' % 'ymax').encode('UTF-8')]
    simzl = sim_dict[('%-80s' % 'zmin').encode('UTF-8')]
    simzr = sim_dict[('%-80s' % 'zmax').encode('UTF-8')]

    xmin = np.asarray(xmin)
    xmax = np.asarray(xmax)

    # Set region to whole simulation box if region is not given.
    if xmin.size == 0 or xmax.size == 0:
        xmin = np.asarray([simxl, simyl, simzl])
        xmax = np.asarray([simxr, simyr, simzr])

    # Checking if the selected box is located within the simulation.
    if np.any(xmax <= np.asarray([simxl, simyl, simzl])) or np.any(xmin >= np.asarray([simxr, simyr, simzr])):
        sys.exit(
            'Selected box is not located within simulation. Please check selected coordinates.\n'
            'Simulation box:\tx\t%1.8e\t%1.8e\n' % (simxl, simxr) +
            '\t\ty\t%1.8e\t%1.8e\n' % (simyl, simyr) +
            '\t\tz\t%1.8e\t%1.8e\n\n' % (simzl, simzr) +
            'Selected box:\tx\t%1.8e\t%1.8e\n' % (xmin[0], xmax[0]) +
            '\t\ty\t%1.8e\t%1.8e\n' % (xmin[1], xmax[1]) +
            '\t\tz\t%1.8e\t%1.8e\n' % (xmin[2], xmax[2])
        )

    # Sanity check coordinates
    if np.any(xmax < xmin):
        sys.exit(
            'Check coordinates. Some value in xmin is larger than xmax.\n' +
            'xmin : %s\t%s\t%s\n' % tuple(xmin) +
            'xmax : %s\t%s\t%s\n' % tuple(xmax)
        )
    if verbose:
        print("minimum position of the region of interest: ", xmin)
        print("maximum position of the region of interest: ", xmax)

    # Find all leaf blocks for inner and outer borders
    # Inner borders only includes blocks which center coordinate is included in selected region.
    # Outer borders also includes blocks which intersect with the selected region. Those variables are denoted with a 2.
    print('Finding all leaf blocks in given region')
    ind = np.argwhere(np.all(gid[:, 7:] == -1, axis=1)).T[0]
    tmpc = coords[ind]
    tmpbs = block_size[ind] * 0.5

    c = np.all(np.concatenate((tmpc >= xmin, tmpc <= xmax), axis=1), axis=1)
    c2 = np.all(np.concatenate((tmpc + tmpbs >= xmin, tmpc - tmpbs <= xmax), axis=1), axis=1)

    blist_raw = ind[c]
    blist_raw2 = ind[c2]

    if len(blist_raw) == 0:
        sys.exit(
            'No block center is availabe in the selected region.'
        )

    print(blist_raw.size, blist_raw2.size)
    bsmin = block_size[ind].min()

    brlvl = refine_level[blist_raw]
    brlvl2 = refine_level[blist_raw2]

    min_ref = brlvl.min()
    min_ref2 = brlvl2.min()
    
    if max_ref_given != None:
        print('Force maximum refinement level: %s' % max_ref_given)
        if max_ref_given < min_ref:
            min_ref = max_ref_given
        if max_ref_given < min_ref2:
            min_ref2 = max_ref_given

    if min_ref_given != None:
        print('Force minimum common refinement level: %s' % min_ref_given)
        if min_ref_given < min_ref:
            min_ref = min_ref_given
        if min_ref_given < min_ref2:
            min_ref2 = min_ref_given

    max_ref_reg = brlvl.max()
    max_ref_reg2 = brlvl2.max()

    max_ref = refine_level[ind].max()
    if verbose:
        print('Biggest block has refinement level: %s' % min_ref)
        print('Highest refinement level in simulation: %s' % max_ref)
        print('Highest refinement level in selected box: %s' % max_ref_reg)

    print('Extending region to fit amr structure at level %s.' % min_ref)
    blist_minref, bx, by, bz, bmax = find_blocks(
        block_list=blist_raw, min_ref_lvl=min_ref, max_ref_lvl=max_ref, block_size=block_size, brlvl=brlvl, bsmin=bsmin,
        coords=coords, gid=gid, refine_level=refine_level, center=(xmin+xmax)/2., is_cuboid=is_cuboid
    )

    while np.any([bx != bmax, by != bmax, bz != bmax]):
        print(
            'Trying extended region given by including blocks whose center is not in predefined region but part '
            'of their block volume is.')

        blist_minref, bx, by, bz, bmax = find_blocks(
            block_list=blist_raw2, min_ref_lvl=min_ref2, max_ref_lvl=max_ref, block_size=block_size, brlvl=brlvl2,
            bsmin=bsmin, coords=coords, gid=gid, refine_level=refine_level, center=(xmin + xmax) / 2., is_cuboid=is_cuboid
        )

        if is_cuboid and type(blist_minref) != int:
            min_ref = min_ref2
            brlvl = brlvl2
            blist_raw = blist_raw2
            max_ref_reg = max_ref_reg2
            break

        if bx == bmax and by == bmax and bz == bmax:
            min_ref = min_ref2
            brlvl = brlvl2
            blist_raw = blist_raw2
            max_ref_reg = max_ref_reg2
            print('New biggest block has refinement level: %s' % min_ref)
            print('New highest refinement level in simulation: %s' % max_ref)
            print('New highest refinement level in selected box: %s' % max_ref_reg)

        else:
            min_ref -= 1
            min_ref2 -= 1
            if min_ref == 0:
                sys.exit('Region could not be extended to a cubic block shape.')

            print('Could not extend region far enough. Extending again at level %s' % min_ref)
            blist_minref, bx, by, bz, bmax = find_blocks(
                block_list=blist_raw, min_ref_lvl=min_ref, max_ref_lvl=max_ref, block_size=block_size, brlvl=brlvl,
                bsmin=bsmin, coords=coords, gid=gid, refine_level=refine_level, center=(xmin+xmax)/2., is_cuboid=is_cuboid
            )

            if is_cuboid and type(blist_minref) != int:
                break

    if is_cuboid:
        blvl = min_ref
    else:
        blvl = min_ref - np.int32(np.round(np.log2(bmax)))
        
    if max_ref_given != None:
        max_ref = max_ref_given

    blist_maxref, b_tot_nr = create_blists(
        minref_blist=blist_minref, block_level=blvl, gid=gid, coords=coords, max_ref_lvl=max_ref,
        bnx=bx, bny=by, bnz=bz, is_cuboid=is_cuboid, verbose=verbose,
    )

    if not is_cuboid:
        b_tot_nr += np.sum(np.logspace(start=0., stop=np.log2(bmax)-1, base=2, num=int(np.log2(bmax)))**3).astype(int)

    if verbose:
        print('Total number of blocks including parent blocks: %s' % b_tot_nr)

        boundariesl = np.min(coords[blist_maxref] - 0.5 * block_size[blist_maxref], axis=0)
        boundariesr = np.max(coords[blist_maxref] + 0.5 * block_size[blist_maxref], axis=0)
        print('Region extended to:\tlower\t\tupper')
        print('\tx\t%+1.8e\t\t%+1.8e' % (boundariesl[0], boundariesr[0]))
        print('\ty\t%+1.8e\t\t%+1.8e' % (boundariesl[1], boundariesr[1]))
        print('\tz\t%+1.8e\t\t%+1.8e\n' % (boundariesl[2], boundariesr[2]))

        print('Region center at:\t%+1.8e\t%+1.8e\t%+1.8e' % tuple((boundariesl+boundariesr)/2))
        print('Given center:\t\t%+1.8e\t%+1.8e\t%+1.8e' % tuple((xmin+xmax)/2))

    # return min & maximum refinement levels in the block, block list, and number of blocks in each direction
    bns = [bx, by, bz]
    brefs = [max_ref, min_ref]

    return blist_maxref, brefs, bns


def get_cube(data, ref_lvl, bbox, bsize, brefs, bns):
    '''
    Transforms the data as a 3-D uniform cube.

    - data: data extracted by h5py filtered by the block list
    - ref_lvl : refinement level for each block, filtered by the block list
    - bbox : the bounding box 
    - bsize : the block size, used to calculate the coordinates
    - brefs : the maximum and minimum refinement level in the region of interest
    - bns : number of blocks in each direction at lowest refinement level
    '''
    # unpack the list
    bmax, bmin = brefs
    bnx, bny, bnz = bns
    
    # maximum level, used to create the uniform grid
    max_lvl = bmax - bmin
    
    # computing the coordiante using the lower & upper corners divided by the smallest block size
    bshape = data.shape[1:4]
    coords = np.round((bbox[:, :, 0] - bbox[0, :, 0]) / bsize.min(axis=0))
    coords = coords.astype(int) * bshape
    
    # shift refinement level by the maximum within the region of interest
    ref_lvl = bmax - ref_lvl
    
    cube = np.zeros((int(bnx) * 2**(max_lvl+3), int(bny) * 2**(max_lvl+3), int(bnz) * 2**(max_lvl+3)), dtype=data.dtype)

    for i in range(len(data)):
        x, y, z = coords[i]
        
        diff = 2**(ref_lvl[i])
        
        data_reshape = np.repeat(data[i], diff, axis=0)
        data_reshape = np.repeat(data_reshape, diff, axis=1)
        data_reshape = np.repeat(data_reshape, diff, axis=2)
        
        sub_cube_size = int(2**(ref_lvl[i] + 3))
        cube[x:x+sub_cube_size, y:y+sub_cube_size, z:z+sub_cube_size] = data_reshape.T
        
    return cube


def get_reflvl_cube(ref_lvl, bbox, bsize, brefs, bns):
    '''
    Returns the refinement level as a 3-D uniform cube.

    - ref_lvl : refinement level for each block, filtered by the block list
    - bbox : the bounding box 
    - bsize : the block size, used to calculate the coordinates
    - brefs : the maximum and minimum refinement level in the region of interest
    - bns : number of blocks in each direction at lowest refinement level
    '''
    # unpack the list
    bmax, bmin = brefs
    bnx, bny, bnz = bns
    
    # maximum level, used to create the uniform grid
    max_lvl = bmax - bmin

    # computing the coordiante using the lower & upper corners divided by the smallest block size
    coords = np.round((bbox[:, :, 0] - bbox[0, :, 0]) / bsize.min(axis=0))
    coords = coords.astype(int) * 8
    
    # shift refinement level by the maximum within the region of interest
    ref_lvl = bmax - ref_lvl

    cube = np.zeros((int(bnx) * 2 ** (max_lvl + 3), int(bny) * 2 ** (max_lvl + 3), int(bnz) * 2 ** (max_lvl + 3)), dtype=ref_lvl.dtype)

    for i in range(len(ref_lvl)):
        x, y, z = coords[i]

        diff = 2 ** (ref_lvl[i]+3)

        data_reshape = np.repeat(ref_lvl[i], diff, axis=0)[..., None]
        data_reshape = np.repeat(data_reshape, diff, axis=1)[..., None]
        data_reshape = np.repeat(data_reshape, diff, axis=2)

        sub_cube_size = int(2 ** (ref_lvl[i]+3))
        cube[x:x + sub_cube_size, y:y + sub_cube_size, z:z + sub_cube_size] = data_reshape.T

    return cube


def get_vector_cube(data_vec, ref_lvl, bbox, bsize, brefs, bns):
    '''
    Wrapper for returning vectorial data as a uniform grid, where the last axis appends the data in each direction.

    - data_vec (list): list of length 3, containing each data in each direction.  extracted by h5py & filtered by the block list
    - bbox : the bounding box 
    - bsize : the block size, used to calculate the coordinates
    - brefs : the maximum and minimum refinement level in the region of interest
    - bns : number of blocks in each direction at lowest refinement level
    '''
    # unpack the list
    bmax, bmin = brefs
    bnx, bny, bnz = bns
    
    # maximum level, used to create the uniform grid
    max_lvl = bmax - bmin

    vec_cube = np.zeros((
        int(bnx) * 2 ** (max_lvl + 3),
        int(bny) * 2 ** (max_lvl + 3),
        int(bnz) * 2 ** (max_lvl + 3),
        3
    ), dtype=data_vec[0].dtype)

    for i in range(3):
        vec_cube[..., i] = get_cube(data_vec[i], ref_lvl, bbox, bsize, brefs, bns)

    return vec_cube


def calc_cube_size(brefs, bns, dtype=np.float32):
    '''
    Calculate the size of the cube in cell units
    
    - brefs : the minimum and maximum refinement level in the region of interest
    - bns : number of blocks in each direction at lowest refinement level
    - dtype: datatype of the cube size. Defaults to float32
    '''
    # unpack the list
    bmax, bmin = brefs
    bnx, bny, bnz = bns
    
    # maximum level, used to create the uniform grid
    max_lvl = bmax - bmin

    # number of cells
    cells = int(bnx) * 2**(max_lvl+3) * int(bny) * 2**(max_lvl+3) * int(bnz) * 2**(max_lvl+3)
    return cells * np.dtype(dtype).itemsize


def get_cdens(data, axis, ref_lvl, bbox, bsize, brefs, bns, weights=None):
    '''
    Get the column density of the data along the axis.

    NB: Only on-axis projections are currently supported.

    - data: data extracted by h5py filtered by the block list
    - axis : the axis in which we want to project [0, 1, or 2]
    - ref_lvl : refinement level for each block, filtered by the block list
    - bbox : the bounding box 
    - bsize : the block size, used to calculate the coordinates
    - brefs : the maximum and minimum refinement level in the region of interest
    - bns : number of blocks in each direction at lowest refinement level
    - weights : optional argument to pass weights for weighted projections. must be the same shape as data. Defaults to None, which performs unweighted projections.
    '''
    # unpack the list
    bmax, bmin = brefs
    
    # maximum level, used to create the uniform grid
    max_lvl = bmax - bmin

    # safety features
    assert axis in [0,1,2], f"Axis {axis} is not 0, 1, or 2 (x, y or z)."

    # remove axis in which projection occrurs
    ax = [0, 1, 2]
    ax.pop(axis)

    bshape = data.shape[1:4]
    coords = np.round((bbox[:, :, 0] - bbox[0, :, 0]) / bsize.min(axis=0))
    coords = coords.astype(int) * bshape

    ref_lvl = bmax - ref_lvl
    sh = data.shape

    use_weights = False
    if weights is not None:
        norm = np.zeros((int(bns[ax[0]]) * 2**(max_lvl+3), int(bns[ax[1]]) * 2**(max_lvl+3)), dtype=data.dtype)

        # make sure shape of weights is same as data
        assert data.shape == weights.shape, "shape of weights " + weights.shape + " != shape of data " + data.shape + " ."

        use_weights = True

    cdens = np.zeros((int(bns[ax[0]]) * 2**(max_lvl+3), int(bns[ax[1]]) * 2**(max_lvl+3)), dtype=data.dtype)

    for i in range(len(data)):
        xyz = coords[i].tolist()

        xyz.pop(axis)

        diff = 2**(ref_lvl[i])
        sub_cube_size = int(2**(ref_lvl[i] + 3))

        if use_weights:
            tmp_data = (data[i] * weights[i]).T.sum(axis=axis)
            weights_reshape = np.repeat(weights[i].T.sum(axis=axis), diff, axis=0)
            weights_reshape = np.repeat(weights_reshape, diff, axis=1)
            norm[xyz[0]:xyz[0]+sub_cube_size, xyz[1]:xyz[1]+sub_cube_size] += weights_reshape
        else:
            tmp_data = data[i].T.sum(axis=axis) * bsize[i, axis] / sh[axis]

        data_reshape = np.repeat(tmp_data, diff, axis=0)
        data_reshape = np.repeat(data_reshape, diff, axis=1)

        cdens[xyz[0]:xyz[0]+sub_cube_size, xyz[1]:xyz[1]+sub_cube_size] += data_reshape
    if use_weights:
        cdens /= norm
    return cdens


def get_slice(data, pos, axis, ref_lvl, bbox, bsize, brefs, bns):
    '''
    Get the slice of the data at the current position.

    - data: data extracted by h5py filtered by the block list
    - pos: 3-D vector of the position we want to slice at
    - axis: axis in which we want to take the slice in
    - ref_lvl : refinement level for each block, filtered by the block list
    - bbox : the bounding box 
    - bsize : the block size, used to calculate the coordinates
    - brefs : the maximum and minimum refinement level in the region of interest
    - bns : number of blocks in each direction at lowest refinement level
    '''

    # unpack the list
    bmax, bmin = brefs
    
    # Get the maximum level difference between base blocks 
    # These don't have to be root blocks of the simulation
    max_lvl = bmax - bmin

    bshape = data.shape[1:4]

    # safety features
    assert axis in [0,1,2], f"Axis {axis} is not 0, 1, or 2 (x, y or z)."

    # As the data and coordinates are stored in different order
    # we need to derive the axis we have for the data insertion seperately
    ax = [0, 1, 2]
    ax_c = ax.pop(axis)
    # Remaining axis are left for expansion if necessary

    # Derive coordinates from the lower corner of the bounding box
    coords = bbox[:, :, 0] - bbox[0, :, 0]
    # Shift the axis such that 0 in the selected axis is our required slice
    # Normalise to smallest block size
    coords /= bsize.min(axis=0)
    # Round values to get next cell position
    coords = np.round(coords)
    # Convert to integer as these are now direct indicies
    coords = coords.astype(int) * bshape
    
    tmp_id = (pos - bbox[:, axis, 0]) / bsize[:, axis] * bshape[axis]
    tmp_id = tmp_id.astype(int)
    # Create array where we store our slice
    sl_ax = np.zeros((int(bns[ax[0]]) * 2**(max_lvl+3), int(bns[ax[1]]) * 2**(max_lvl+3)))
    # Shift the refinement level such that it represent the differen to the highest level
    ref_lvl = bmax - ref_lvl

    # Loop over all blocks
    for i in range(len(data)):
        # Get indicies for current lower corner of the block
        xyz = coords[i].tolist()
        # Get the lower index along the slice axis
        pi = xyz.pop(axis)

        # Get factor of block size in relation to smallest block
        diff = 2**(ref_lvl[i])
        
        # Check if our 0 index is within the range of the current block
        if tmp_id[i] < 0 or tmp_id[i] >= bshape[axis]:
            continue
        
        # Create general slice array
        sel = [slice(None, None, None), slice(None, None, None), slice(None, None, None)]
        sel[ax_c] = slice(tmp_id[i], tmp_id[i]+1, None)

        # Increase block size to fit the refinement level
        data_reshape = np.repeat(data[i].T[sel[0], sel[1], sel[2]], diff, axis=ax[0])
        data_reshape = np.repeat(data_reshape, diff, axis=ax[1])
        # Get size of the slice patch
        sub_cube_size = int(2**(ref_lvl[i] + 3))
        # Store in slice array, needs to be transposed to convert from ZYX to XYZ
        sl_ax[xyz[0]:xyz[0]+sub_cube_size, xyz[1]:xyz[1]+sub_cube_size] = np.squeeze(data_reshape)

    return sl_ax



'''
Supplementary module to contain low-level routines for block determination
'''
import numpy as np
from .zorder import zenumerate

# Find all blocks at lowest refinement level which are correspoding to our initial guess
def find_blocks(
        block_list, min_ref_lvl, max_ref_lvl, brlvl, coords, block_size, bsmin, refine_level, gid, center, is_cuboid=False):

    # For all block which are higher than our set refinement level find the lowest parent which is fulfills our
    # requested refinement level.
    blist_ref = []
    for block in block_list[brlvl > min_ref_lvl]:
        b_tmp = block
        while refine_level[b_tmp] > min_ref_lvl:
            b_tmp = gid[b_tmp][6] - 1
        blist_ref.append(b_tmp)

    # As there will be many overlaps we create a unique list of block ids
    blist_ref = np.unique(blist_ref)
    # And add them to all the cells which already were at the lowest refinement level.
    minref_blist = np.unique(np.concatenate((block_list[brlvl == min_ref_lvl], blist_ref))).astype(int)

    # Check how many blocks fit into the current range in x, y and z at the lowest refinement level.
    bnx = np.round((np.amax(
        coords[minref_blist, 0] + 0.5 * block_size[minref_blist, 0])
        - np.amin(coords[minref_blist, 0] - 0.5 * block_size[minref_blist, 0])
        )/(bsmin * 2**(max_ref_lvl-min_ref_lvl)))

    bny = np.round((np.amax(
        coords[minref_blist, 1] + 0.5 * block_size[minref_blist, 1])
        - np.amin(coords[minref_blist, 1] - 0.5 * block_size[minref_blist, 1])
        )/(bsmin * 2**(max_ref_lvl-min_ref_lvl)))

    bnz = np.round((np.amax(
        coords[minref_blist, 2] + 0.5 * block_size[minref_blist, 2])
        - np.amin(coords[minref_blist, 2] - 0.5 * block_size[minref_blist, 2])
        )/(bsmin * 2**(max_ref_lvl-min_ref_lvl)))

    # Calculate the max number of blocks at lowest refinement level needed to be able to create an amr tree.
    bn = np.max([bnx, bny, bnz])
    bnmax = 2**np.ceil(np.log2(bn))

    # Check if we have all our blocks fill the cuboid at least in numbers. Let's hope this is never the case.
    # Currently no idea how to fix this.
    if not bnx * bny * bnz == minref_blist.shape[0]:
        print('Blocks in x: ', bnx)
        print('Blocks in y: ', bny)
        print('Blocks in z: ', bnz)
        print(minref_blist.shape)
        return 1, 1, 1, 1, 2
        # sys.exit('Could not find enough blocks to fill cuboid.')
    elif is_cuboid:
        return minref_blist, bnx, bny, bnz, bnmax

    # For each round add blocks at at least one side to reach the maximum number of blocks for amr tree.
    for i in range(int(bnmax - np.min([bnx, bny, bnz]) + 2)):

        minref_blist, bnx, bny, bnz = add_blocks(
            minref_blist=minref_blist, coords=coords, block_size=block_size, gid=gid, bnx=bnx, bny=bny, bnz=bnz,
            bnmax=bnmax, center=center, axis=0)

        minref_blist, bnx, bny, bnz = add_blocks(
            minref_blist=minref_blist, coords=coords, block_size=block_size, gid=gid, bnx=bnx, bny=bny, bnz=bnz,
            bnmax=bnmax, center=center, axis=1)

        minref_blist, bnx, bny, bnz = add_blocks(
            minref_blist=minref_blist, coords=coords, block_size=block_size, gid=gid, bnx=bnx, bny=bny, bnz=bnz,
            bnmax=bnmax, center=center, axis=2)

    return minref_blist, bnx, bny, bnz, bnmax


def add_blocks(minref_blist, coords, block_size, gid, bnx, bny, bnz, bnmax, center, axis):

    bn = np.asarray([bnx, bny, bnz])
    gid_ind = 2 * axis

    # Check if there are neghbours availabe for all blocks in + and - direction of the given axis.
    nb_avail_p = np.all(
        coords[
            gid[minref_blist[coords[minref_blist, axis] == coords[minref_blist, axis].max()], gid_ind + 1] - 1, axis
        ] > coords[minref_blist, axis].max()
    )

    if nb_avail_p:
        if np.all(gid[minref_blist[coords[minref_blist, axis] == coords[minref_blist, axis].max()], gid_ind + 1] - 1 < 0):
            nb_avail_p = False

    nb_avail_m = np.all(
        coords[
            gid[minref_blist[coords[minref_blist, axis] == coords[minref_blist, axis].min()], gid_ind] - 1, axis
        ] < coords[minref_blist, axis].min()
    )

    if nb_avail_m:
        if np.all(gid[minref_blist[coords[minref_blist, axis] == coords[minref_blist, axis].min()], gid_ind] - 1 < 0):
            nb_avail_m = False

    new_b_p = []
    new_b_m = []
    # Check if the maximum number of blocks at lowest ref. level are reached and if the difference greater than 1.
    # In that case we add blocks at either side, if available, otherwise we only add blocks at one side and check
    # which keeps the box center the closest to the given center.
    if bn[axis] != bnmax and bnmax - bn[axis] >= 2:
        if nb_avail_p:
            new_b_p = gid[minref_blist[coords[minref_blist, axis] == coords[minref_blist, axis].max()], gid_ind + 1] - 1
            new_b_p = new_b_p[new_b_p > 0]
            if new_b_p.size == int(np.prod(bn)/bn[axis]):
                bn[axis] += 1
            else:
                new_b_p = []
        if nb_avail_m:
            new_b_m = gid[minref_blist[coords[minref_blist, axis] == coords[minref_blist, axis].min()], gid_ind] - 1
            new_b_m = new_b_m[new_b_m > 0]
            if new_b_m.size == int(np.prod(bn)/bn[axis]):
                bn[axis] += 1
            else:
                new_b_m = []
    elif bn[axis] != bnmax:
        if nb_avail_p and nb_avail_m:
            trig = False
            new_b_p = gid[minref_blist[coords[minref_blist, axis] == coords[minref_blist, axis].max()], gid_ind + 1] - 1
            new_b_p = new_b_p[new_b_p > 0]
            new_b_m = gid[minref_blist[coords[minref_blist, axis] == coords[minref_blist, axis].min()], gid_ind] - 1
            new_b_m = new_b_m[new_b_m > 0]
            boundariesl = np.min(coords[new_b_m] - 0.5 * block_size[new_b_m], axis=axis)[axis]
            boundariesr = np.max(coords[new_b_p] + 0.5 * block_size[new_b_p], axis=axis)[axis]
            if np.abs(boundariesl - center[axis]) < np.abs(boundariesr - center[axis]):
                if new_b_m.size == int(np.prod(bn)/bn[axis]):
                    new_b_p = []
                    bn[axis] += 1
                else:
                    new_b_m = []
                    trig = True
            if np.abs(boundariesl - center[axis]) >= np.abs(boundariesr - center[axis]) or trig:
                if new_b_p.shape[axis] == int(np.prod(bn)/bn[axis]):
                    new_b_m = []
                    bn[axis] += 1
                else:
                    new_b_p = []
                    new_b_m = []
        elif nb_avail_p:
            new_b_p = gid[minref_blist[coords[minref_blist, axis] == coords[minref_blist, axis].max()], gid_ind + 1] - 1
            new_b_p = new_b_p[new_b_p > 0]
            if new_b_p.size == int(np.prod(bn)/bn[axis]):
                bn[axis] += 1
            else:
                new_b_p = []
        elif nb_avail_m:
            new_b_m = gid[minref_blist[coords[minref_blist, axis] == coords[minref_blist, axis].min()], gid_ind] - 1
            new_b_m = new_b_m[new_b_m > 0]
            if new_b_m.size == int(np.prod(bn)/bn[axis]):
                bn[axis] += 1
            else:
                new_b_m = []

    # Add new blocks to the current list of blocks at lowest refinement level.
    minref_blist = np.unique(np.concatenate((minref_blist, new_b_p, new_b_m))).astype(int)

    return minref_blist, bn[0], bn[1], bn[2]


# Sort all blocks at minimum refinement level and replace block with their higher refinement level counterparts.
def create_blists(minref_blist, max_ref_lvl, block_level, gid, coords, bnx=0, bny=0, bnz=0, is_cuboid=False, verbose=False):

    # Put the block of the minimum refinement level on a grid correspoding to their coordinates.
    # Change axes to be in order of x, y and z.
    blist_minsort_tmp = blocks_on_grid(b_ids=minref_blist, coords=coords, bnx=bnx, bny=bny, bnz=bnz).swapaxes(0, 2)
    if verbose:
        print('bnx, bny, bnz: ', bnx, bny, bnz)
        print('Block shape: ', blist_minsort_tmp.shape)
    blist_minsort = []
    # Add blocks in amr order to list.
    if is_cuboid:
        blist_minsort = blist_minsort_tmp.swapaxes(0, 2).flatten()
    else:
        for pos in zenumerate(blist_minsort_tmp.shape):
            blist_minsort.append(blist_minsort_tmp[pos])

    # Check all blocks if they have children (!-1) using the gid and replace the corresponding block with them.
    maxref_blist = np.asarray(blist_minsort)
    gid_tmp = gid[maxref_blist, 7:]
    tot_nr_blocks = maxref_blist.size
    # Goes through all refinement levels one by one.
    for j in range(max_ref_lvl - block_level):
        # Adds from back to front to circumvent changing the position every time blocks are added.
        for k in reversed(range(len(gid_tmp))):
            if gid_tmp[k, 0] != -1:
                maxref_blist = np.delete(maxref_blist, k)
                maxref_blist = np.insert(maxref_blist, k, gid_tmp[k]-1)
                tot_nr_blocks += 8
        gid_tmp = gid[maxref_blist, 7:]
    return maxref_blist, tot_nr_blocks
    

def blocks_on_grid(b_ids, coords, bnx=0, bny=0, bnz=0):

    if bnx == 0 or bny == 0 or bnz == 0:
        bn = 2**(int(np.log2(len(b_ids)))/3)
        bnx = bn
        bny = bn
        bnz = bn

    bgrid = np.reshape(b_ids[np.argsort(a=coords[b_ids, 2])], (int(bnz), int(bnx * bny)))

    for i in range(int(bnz)):
        bgrid[i] = bgrid[i, np.argsort(coords[bgrid[i]][:, 1])]
    bgrid = bgrid.reshape((int(bnz), int(bny), int(bnx)))

    for i in range(int(bnz)):
        for j in range(int(bny)):
            bgrid[i, j] = bgrid[i, j, np.argsort(coords[bgrid[i, j]][:, 0])]

    return bgrid


################################################################################
# Copyright (c) 2012, Lawrence Livermore National Security, LLC.
# Produced at the Lawrence Livermore National Laboratory
# Written by Todd Gamblin et al. <tgamblin@llnl.gov>
# LLNL-CODE-599252
# All rights reserved.
#
# This file is part of Rubik. For details, see http://scalability.llnl.gov.
# Please read the LICENSE file for further information.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the disclaimer below.
#
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the disclaimer (as noted below) in the
#       documentation and/or other materials provided with the distribution.
#
#     * Neither the name of the LLNS/LLNL nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL LAWRENCE LIVERMORE NATIONAL SECURITY, LLC, THE
# U.S. DEPARTMENT OF ENERGY OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
# OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
################################################################################
"""
This file provides routines to transform the elements of an ndarray from
dimension-major order to Z-order.
"""

import numpy as np
import itertools
import math


def le_power_of_2(num):
    """ Gets a power of two less than or equal to num. Works for up to 64-bit
    numbers.
    """
    num |= (num >> 1)
    num |= (num >> 2)
    num |= (num >> 4)
    num |= (num >> 8)
    num |= (num >> 16)
    num |= (num >> 32)
    return num - (num >> 1)


def b(num, bytes=8):
    bytefields = ["{0:08b}".format(num >> 8 * i & 0xFF) for i in reversed(range(bytes))]
    return " ".join(bytefields).replace('0', '-')


def h(num, bytes=8):
    quads = bytes / 2
    quadfields = ["{0:04x}".format(num >> 8 * i & 0xFFFF) for i in reversed(range(quads))]
    return " ".join(quadfields).replace('0', '-')


class ZEncoder(object):
    """ Class representing a set of bitmasks for encoding/decoding
    n-dimensional morton numbers.
    Parameters:
    ndim    number of dimensions to encode/decode for.
    bits    number of bits in the generated codes.
    Note: Codes are internally generated with as many bits as are necessary,
    then they are returned as either numpy.uint32 or numpy.uint64, depending
    on how many bits are needed to represent the codes.
    """
    filters = {}

    def __init__(self, ndim, bits=64):
        self.ndim = ndim
        self.bits = bits
        self.filter = self.get_filter(ndim, bits)

    @classmethod
    def for_shape(cls, shape):
        # Default to 32-bit codes, but use 64-bit codes if needed for a shape.
        maxdim = max(shape)
        if math.ceil(math.log(maxdim, 2)) >= (32 / len(shape)):
            bits = 64
        else:
            bits = 32
        return ZEncoder(len(shape), bits)

    def get_filter(self, ndim, bits):
        """ Get a possibly memoized filter for the dimensions bit length
            specified.
        """
        key = (ndim, bits)
        filters = self.__class__.filters
        if not key in filters:
            filters[key] = ZEncoder.create_filter(ndim, bits)
        return filters[key]

    @classmethod
    def create_filter(cls, ndim, bits):
        """ This creates a filter (a set of bitmasks) that can be used to
        quickly generate n-dimensional Z codes (Morton codes). The filter is
        based on the method described here:
        http://graphics.stanford.edu/~seander/bithacks.html#InterleaveBMN
            Parameters:
        ndim
          number of dimensions in the z curve
            bits
          bit width of morton codes to be generated. Each coordinate gets
          bits/ndim bits.
            """
        # Construct the initial mask: this selects just the lower (bits/ndim)
        # bits of a number
        width = bits // ndim
        mask = (1 << width) - 1

        # Each bit in a coordinate needs to move (ndim * width) positions.
        # We'll move them using iterative power-of-2 shifts for O(log(bits))
        # complexity.
        # First get the max shift we need to do:
        max_shift = le_power_of_2(ndim * (width - 1))
        filter = [[mask, 0, max_shift]]

        # Now figure out which bits need to be moved by each shift, and build
        # masks.
        shift = max_shift
        while shift > 0:
            mask = 0
            shifted = 0
            shift_mask = ~(shift - 1)
            for bit in range(width):
                distance = ndim * bit - bit
                shifted |= (shift & distance)
                mask |= 1 << bit << (distance & shift_mask)
            if shifted:
                filter.append([mask, shift, shift >> 1])
            shift >>= 1

        # set last rshift to zero for compact operation.
        filter[-1][2] = 0
        return filter

    def spread(self, x):
        """ Applies filter to spread the bits of x apart by ndim-1 zeros. """
        for mask, shift, rshift in self.filter:
            x = (x | (x << shift)) & mask
        return x

    def compact(self, x):
        """ Applies filter in reverse to push spread bits back together. """
        for mask, shift, rshift in self.filter[-1::-1]:
            x = (x | (x >> rshift)) & mask
        return x

    def encode(self, point):
        """ Takes a point and returns a morton code for that point. """
        if len(point) != self.ndim:
            raise Exception(
                "Error: Can't encode %d-dimensional point with %d-dimensional ZEncoder." % (len(point), self.ndim))
        code = 0
        for i, x in enumerate(point):
            code |= self.spread(x) << i

        # Python will do arithmetic with the largest ints it can.
        # This ensures we return an int32 when it was asked for.
        return code

    def decode(self, value):
        """ Given an ndim-dimensional morton code, returns the corresponding
        point as a tuple.
        """
        return tuple([self.compact(value >> i) for i in range(self.ndim)])

    def __bytes(self):
        """ Number of bytes needed to represent the masks in this filter. """
        if self.bits > 32:
            return 8
        else:
            return 4

    def __str(self, format):
        """ Prints out each mask in the filter line along with its
        corresponding left and right shifts. format paramter determines how to
        format binary numbers. Options are b or h.
            """
        fields = ["%s %2d %2d" % (format(mask, self.__bytes()), shift, rshift) for mask, shift, rshift in self.filter]
        return "\n".join(fields)

    def hex_str(self):
        """ formatted string with masks in the filter in hexadecimal, along
        with their left and right shifts.
        """
        return self.__str(h)

    def bin_str(self):
        """ formatted string with masks in the filter in binary, along with
        their left and right shifts.
        """
        return self.__str(b)

    def __str__(self):
        """ Equivalent to bin_str() """
        return self.bin_str()


def zenumerate(shape):
    """ Enumerates points in the shape in Z order. Currently dumps the morton
    codes into an array and sorts them, then regenerates points in that
    order. This is O(nlogn) time. We could do better for matrices with more
    even aspect ratios by enuerating all morton codes from 0 on and
    converting to points, but that is O(n^2) for irregular shapes.
    """
    # Build a buffer of encoded z values and sort them
    zencoder = ZEncoder.for_shape(shape)
    buffer = [zencoder.encode(point) for point in np.ndindex(*shape)]
    buffer.sort()

    # Decode z values in order and yield each.
    for code in buffer:
        yield zencoder.decode(code)


def zorder(arr):
    """ Transform the elements of an ndarray from dimension-major order to z
    order. This modifies the array.
    """
    buffer = arr.copy()
    i = 0
    for index in zenumerate(arr.shape):
        arr[index] = buffer.flat[i]
        i += 1
    return arr